# -*- coding: UTF-8 -*-

# Phatch - Photo Batch Processor
# Copyright (C) 2007-2008 www.stani.be
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/
#
# Phatch recommends SPE (http://pythonide.stani.be) for python editing.

# Follows PEP8

# import rpdb2; rpdb2.start_embedded_debugger('x')

import datetime
import math
import os
import random
import re
import types
from translation import _t
from lib import unicoding

RE_VAR = re.compile('<.+?>')
RE_PY_VAR = \
    re.compile('(?P<value>[A-Za-z]\w*)([.]\w(\w|[.])+)?|(?P<format>#+)')
RE_FORMAT = re.compile('^#+$')
_MODULE = '(^|<)%s[.]'
MODULES = {
    'EXIF': re.compile(_MODULE%'EXIF'),
    'pyexiv2': re.compile(_MODULE%'(Exif|Iptc)'),
    'Pexif': re.compile(_MODULE%'Pexif'),
}

DATETIME_KEYS = ['year', 'month', 'day', 'hour', 'minute', 'second']
RE_DATETIME_KEYS = re.compile('(%s)$'%('|'.join(['[.]'+key
    for key in DATETIME_KEYS])))
MONTHS = (_t('January'), _t('February'), _t('March'), _t('April'),
    _t('May'), _t('June'), _t('July'), _t('August'), _t('September'),
    _t('October'), _t('November'), _t('December'))
WEEKDAYS = (_t('Monday'), _t('Tuesday'), _t('Wednesday'), _t('Thursday'),
    _t('Friday'), _t('Saturday'), _t('Sunday'))


def _add_module(d, module_dict):
    if not(type(module_dict) is dict):
        module_dict = module_dict.__dict__
    for key, value in module_dict.items():
        if key[0] != '_':
            d[key] = value


def is_string(x):
    return type(x) in types.StringTypes


def get_vars(actions):
    """Extract all used variables from actions.

    :param actions: list of actions
    :type actions: list of dict
    """
    vars = []
    for action in actions:
        fields = action['fields']
        for value in fields.values():
            for var in RE_VAR.findall(value):
                var = var[1:-1] #strip <>
                if not var in vars:
                    vars.append(var)


def get_module(var):
    for module, regex in MODULES.items():
        if regex.search(var):
            return module
    return None


def get_modules(vars):
    modules = []
    for var in vars:
        module = get_module(var)
        if module and not(module in modules):
            modules.append(module)
    return modules

RE_DATETIME_COLON = \
re.compile('[0-9]{4,4}:[0-9]{2,2}:[0-9]{2,2} [0-9]{2,2}:[0-9]{2,2}:[0-9]{2,2}')


class InvalidDateTimeError(Exception):
    pass


class InfoCache(object):
    type = 'cache'

    def __init__(self, source=None):
        """Creates a cached info. The base is the source from which
        all variables are extracted on demand. When all variables
        have to be extracted use the :method:``extract_all`` method.
        The source can be a static object or a callable for cases
        where the object is manipulated and the variables remain
        valid (e.g. when resizing an image the mode stays valid)

        The difference between ``get`` and ``extract`` is that ``get``
        returns a value, while ``extract`` extracts it and places it in
        ``dict``. ``extract`` allows for multiple, related values to
        be extracted simultaneously.

        :param source: retrieve source from which to extract data
        :type source: object or callable
        """
        self.prefix = self.type+'_'
        self._prefix_n = len(self.prefix)
        self._set_extract_methods()
        self.vars = sorted(self._extract_methods.keys())
        if not (source is None):
            self.set_source(source)

    def provides(self, var):
        """Wether this info (might) provide the ``var`` variable.

        :returns: if this info should handle this variable
        :rtype: bool
        """
        return var in self.vars or var[:self._prefix_n] == self.prefix

    def set_source(self, source):
        """Sets the source of this info. This allows reuse of the same
        info if needed.

        :param source: retrieve source from which to extract data
        :type source: object or callable

        .. see also: :method:`_set_source`
        """
        if callable(source):
            self._get_source_dynamic = source
            self._update_source = self._update_callable_source
            source = None
        else:
            self._get_source_dynamic = None
            self._update_source = self._update_static_source
        self._set_source(source)
        self.disable_cache()
        self.dict = {}

    def _set_source(self, source):
        """Sets the source at the lowest level. Helper method for
        :method:`set_source` and :method:`_update_callable_source`.

        :param source: retrieve source from which to extract data
        :type source: object or callable
        """
        self._source = source

    def _set_extract_methods(self):
        """To be overwritten if the info has preknown varialbes."""
        self._extract_methods = {}

    def enable_cache(self):
        """Enable the static cache so that the expensive
        :method:``_get_source_dynamic`` does not get called for
        every var access.

        This makes only sense for callable sources.
        """
        self._update_source()
        type(self).__getitem__ = self._get_var_from_source_without_update

    def disable_cache(self):
        """Turn off static caching.

        This makes only sense for callable sources.
        """
        type(self).__getitem__ = self._get_var_from_source_with_update

    def _update_static_source(self):
        """Update static cache does not do anything. This is just
        a dummy function which can be assigned to
        :method:``_update_source`` if the source is not callable."""
        pass

    def _update_callable_source(self):
        """Update dynamic cache updates ``self._source`` with
        ``self._get_source_dynamic``. This can be assigned to
        :method:``_update_source`` if the source is callable.

        .. see also: :method:`_set_source`
        """
        self._set_source(self._get_source_dynamic())

    def _get_var_from_source_with_update(self, var):
        """If the variable ``var`` is not present in the ``dict``,
        this method updates the ``source`` first and afterwards
        extracts the variable ``var`` from source.

        :param var: variable
        :type var: string
        """
        try:
            return self.dict[var]
        except KeyError:
            self._update_source()
            return self._get_var_from_source(var)

    def _get_var_from_source_without_update(self, var):
        """If the variable ``var`` is not present in the ``dict``,
        this extracts the variable ``var`` from source without updating
        the ``source`` (as it is static).

        :param var: variable
        :type var: string
        """
        try:
            return self.dict[var]
        except KeyError:
            return self._get_var_from_source(var)

    def _get_var_from_source(self, var):
        """

        :param var: variable
        :type var: string
        """
        self._extract_var_with_method_or_other(var)
        return self.dict[var]

    def _extract_var_with_method_or_other(self, var):
        """Extract var and store it in a dictionary.

        :param var: variable
        :type var: string
        """
        if var in self._extract_methods:
            self._extract_methods[var]()
        else:
            self._extract_other_method(var)

    def _extract_other_method(self, var):
        """Extract only one var"""
        if var[:self._prefix_n] == self.prefix:
            self.dict[var] = self._get_other(var[self._prefix_n:])
        else:
            raise KeyError(var)

    def _extract_other(self, var):
        """Extract only one other ``var`` into ``dict``. This raises a
        ``KeyError`` by default. Helper function for
        :method:`_extract_other_method`. (to be overwritten)

        :param var: variable
        :type var: string
        """
        raise KeyError(self.prefix+var)

    def extract_others(self):
        """Extract all other vars. Does nothing by default.
        (to be overwritten)"""

    def extract_all(self):
        """Extract all values, which is usefull for inspector."""
        for var, extract in self._extract_methods.items():
            if not(var in self.dict):
                extract()
        self.extract_others()


class UnknownTypeError(Exception):
    pass


class InfoFile(InfoCache):
    """Wraps a lazy file path access around an image filename.

    >>> info = InfoFile('/home/phatch/test.png')
    >>> info['foldername']
    u'phatch'
    >>> sorted(info.dict.keys())
    ['foldername', 'root']
    >>> info['type']
    u'png'
    >>> sorted(info.dict.keys())
    ['filename', 'foldername', 'root', 'type']
    >>> info.set_source('/home/gimp/world.jpg')
    >>> info['type']
    u'jpg'
    >>> sorted(info.dict.keys())
    ['filename', 'type']
    """

    type = 'File'

    def _set_extract_methods(self):
        self._extract_methods = {
            _t('day'): self._extract_stat,
            _t('foldername'): self._extract_root,
            _t('filename'): self. _extract_filename,
            _t('filesize'): self._extract_stat,
            _t('folder'): self._extract_folder,
            _t('hour'): self._extract_stat,
            _t('minute'): self._extract_stat,
            _t('month'): self._extract_stat,
            _t('monthname'): self._extract_stat,
            _t('path'): self._extract_path,
            _t('root'): self._extract_root,
            _t('second'): self._extract_stat,
            _t('type'): self. _extract_filename,
            _t('weekday'): self._extract_stat,
            _t('weekdayname'): self._extract_stat,
            _t('year'): self._extract_stat,
            }

    def _set_source(self, source):
        if is_string(source):
            self._source_path = source
            self._source_parent = os.path.dirname(self._source_path)
        else:
            self._source_path, self._source_parent = source
        self._source_path = unicoding.ensure_unicode(self._source_path)
        self._source_parent = unicoding.ensure_unicode(self._source_parent)

    def _extract_path(self):
        self.dict['path'] = self._source_path

    def _extract_folder(self):
        self.dict['folder'] = self._source_parent

    def _extract_root(self):
        self.dict['root'], self.dict['foldername'] = \
            os.path.split(self._source_parent)

    def _extract_filename(self):
        filename = os.path.basename(self._source_path)
        self.dict['filename'], typ = \
            os.path.splitext(filename)
        self.dict['type'] = typ[1:]

    def _extract_stat(self):
        #TODO: PORT THIS TO DateTime
        #os time
        try:
            file_stat = os.stat(self._source_path)
            self.dict['filesize'] = file_stat.st_size
            st_mtime = time.localtime(file_stat.st_mtime)[:7]
        except:
            self.dict['filesize'] = 0
            st_mtime = (0, )*7
        self.dict['year'], self.dict['month'], self.dict['day'], \
            self.dict['hour'], self.dict['minute'], self.dict['second'], \
            self.dict['weekday'] = st_mtime
        self.dict['weekdayname'] = WEEKDAYS[self.dict['weekday']]
        self.dict['monthname'] = MONTHS[self.dict['month']-1]


class InfoPil(InfoCache):
    """Wraps a lazy PIL var access to an image.

    :param image: Pil.Image or callable to retrieve it
    :type image: Pil.Image/function

    >>> import pprint
    >>> import Image
    >>> image = Image.new('L',(1,2))
    >>> info = InfoPil(image)
    >>> info.provides('Pil_FormatDescription')
    True
    >>> pprint.pprint(info.vars)
    ['Pil_Format',
     'Pil_FormatDescription',
     'Pil_Size',
     'compression',
     'height',
     'mode',
     'width']
    >>> sorted(info.dict.keys())
    []
    >>> info['mode']
    'L'
    >>> info['height']
    2
    >>> info['Pil_Format']
    >>> sorted(info.dict.keys())
    ['Pil_Format', 'Pil_Size', 'height', 'mode', 'width']
    >>> info.reset_geometry()
    >>> sorted(info.dict.keys())
    ['Pil_Format', 'Pil_Size', 'mode']
    >>> info.extract_all()
    >>> pprint.pprint(sorted(info.dict.keys()))
    ['Pil_Format',
     'Pil_FormatDescription',
     'Pil_Size',
     'compression',
     'height',
     'mode',
     'width']
    """
    type = 'Pil'

    def _set_extract_methods(self):
        self._extract_methods = {
            _t('compression'): self._extract_compression,
            _t('height'): self._extract_size,
            _t('mode'): self._extract_mode,
            _t('width'): self._extract_size,
            'Pil_Format': self._extract_format,
            'Pil_FormatDescription': self._extract_format_description,
            'Pil_Size': self._extract_size}

    def _extract_compression(self):
        self.dict['compression'] = \
            self._source.info.get('compression', 'none')

    def _extract_format(self):
        self.dict['Pil_Format'] = self._source.format

    def _extract_format_description(self):
        self.dict['Pil_FormatDescription'] = \
            self._source.format_description

    def _extract_mode(self):
        self.dict['mode'] = self._source.mode

    def _extract_size(self):
        """Extract size of image with PIL"""
        #we don't translate dotted attributes
        self.dict['width'], self.dict['height'] = \
            self.dict['Pil_Size'] = self._source.size

    def _extract_transparency(self):
        self.dict['transparency'] = \
            self._source.info.get('transparency', None)

    def _get_other(self, var):
        """Extract only one var"""
        try:
            return self._source.info[var]
        except KeyError:
            raise KeyError(self.prefix+var)

    def extract_others(self):
        """Extract all other vars"""
        for key, value in self._source.info.items():
            if key!=self.prefix:
                self.dict[self.prefix+key.title()] = value

    def reset_geometry(self):
        for var in ('width', 'height', 'Pil_size'):
            if var in self.dict:
                del self.dict[var]


def extract_info_pexif(info, image):
    """pexif = Pil EXIF"""
    #check if exif information is available through pil
    if not hasattr(image, '_getexif'):
        return info
    #add exif information
    try:
        exif = image._getexif()
    except:
        exif = None
    if not exif:
        return
    for key, value in exif.items():
        new = {'Zexif%s0x%04x'%(DOT, key): value}
        if key in EXIFTAGS:
            new['Pexif%s%s'%(DOT, EXIFTAGS[key])] = value
        #Check if we can split exif data further
        split_data(new)
        info.update(new)


try:
    from ExifTags import TAGS, GPSTAGS
    EXIFTAGS = {}
    EXIFTAGS.update(TAGS)
    EXIFTAGS.update(GPSTAGS)
    del TAGS
    del GPSTAGS
except:
    #older versions of PIL
    EXIFTAGS = {}
EXIFTAGS_REVERSE = {}
for key, item in EXIFTAGS.items():
    EXIFTAGS_REVERSE[item] = key


def datetime_from_exif(exif_string):
    d = datetime.datetime(1, 1, 1)
    return d.strptime(exif_string, '%Y:%m:%d %H:%M:%S')


class DateTime:
    format = '%Y:%m:%d %H:%M:%S'

    def __init__(self, exif_string):
        self.from_exif(exif_string)

    def __getattr__(self, attr):
        if attr == 'weekday':
            return self.datetime.weekday()
        elif attr == 'weekdayname':
            return WEEKDAYS[self.datetime.weekday()]
        elif attr == 'monthname':
            return MONTHS[self.datetime.month]
        else:
            return getattr(self.datetime, attr)

    def from_exif(self, exif_string):
        d = datetime.datetime(1, 1, 1)
        self.datetime = d.strptime(exif_string, self.format)

    def __str__(self):
        return self.datetime.strftime(self.format)

    def __repr__(self):
        return "DateTime('%s')"%self


class InfoPexif(InfoCache):
    """Wraps a lazy PIL exif var access to an image.

    >>> import os
    >>> import Image
    >>> image = Image.open('../tests/images/input/exif.jpg')
    >>> info = InfoPexif(image)
    >>> info['Pexif_DateTimeOriginal']
    DateTime('2009:09:02 13:15:17')
    """
    type = 'Pexif'

    def provides(self, var):
        return var[:self._prefix_n] == self.prefix and \
            var[self._prefix_n:] in EXIFTAGS_REVERSE

    def _get_other(self, var):
        """Extract only one other ``var``

        :param var: variable
        :type var: string
        """
        #import rpdb2; rpdb2.start_embedded_debugger('x')
        try:
            key = EXIFTAGS_REVERSE[var]
        except KeyError:
            #re raise: give full name to key error
            raise KeyError(self.prefix+var)
        try:
            exif = self._source._getexif()
        except:
            raise KeyError(self.prefix+var)
        value = exif[key]
        if RE_DATETIME_COLON.match(value):
            value = DateTime(value)
        return value


class InfoFormat(InfoCache):
    type = 'Format'

    def __init__(self):
        super(InfoFormat, self).__init__()
        self.set_source(None)

    def provides(self, var):
        return RE_FORMAT.match(var)

    def _extract_other_method(self, var):
        """Extract only one var"""
        match = RE_FORMAT.match(var)
        if match:
            self.dict[var] = '"%%0%dd"%%'%len(var)
        else:
            raise KeyError(var)

GLOBALS = {}
_add_module(GLOBALS, math)
_add_module(GLOBALS, random)
_add_module(GLOBALS, __builtins__)
del GLOBALS['copyright']
GLOBALS['DateTime'] = DateTime


class InfoEval:
    """Create an info like dictionary which uses a collection of several
    info instances and can evaluate Python expressions.

    >>> import pprint
    >>> import Image
    >>> filename = '../tests/images/input/exif.jpg'
    >>> image1 = Image.open(filename)
    >>> image2 = Image.new('L',(1,2))
    >>> info = InfoEval(locals={'index':2, 'folderindex':3})
    >>> info.list.append(InfoFormat())
    >>> info.list.append(InfoFile(filename))
    >>> info.list.append(InfoPil(image1))
    >>> info.list.append(InfoPexif(image1))
    >>> info.types()
    ['Format', 'File', 'Pil', 'Pexif']
    >>> sorted(info.dump().items())
    [('folderindex', 3), ('index', 2)]
    >>> info['Pil_Format']
    'JPEG'
    >>> sorted(info.dump().items())
    [('Pil_Format', 'JPEG'), ('folderindex', 3), ('index', 2)]
    >>> info['width']
    640
    >>> info['####width']
    '0640'
    >>> info['width+pi']
    643.14159265358978
    >>> info['mode.lower()']
    'rgb'
    >>> info['random()'] == info['random()']
    False
    >>> info['Pexif_DateTimeOriginal.year']
    2009
    >>> info['Pexif_DateTimeOriginal.weekdayname.upper()']
    'WEDNESDAY'
    >>> info['weekdayname']
    'Monday'
    >>> pprint.pprint(sorted(info.dump().items()))
    [('####', '"%04d"%'),
     ('Pexif_DateTimeOriginal', DateTime('2009:09:02 13:15:17')),
     ('Pil_Format', 'JPEG'),
     ('Pil_Size', (640, 480)),
     ('day', 0),
     ('filesize', 0),
     ('folderindex', 3),
     ('height', 480),
     ('hour', 0),
     ('index', 2),
     ('minute', 0),
     ('mode', 'RGB'),
     ('month', 0),
     ('monthname', 'December'),
     ('second', 0),
     ('weekday', 0),
     ('weekdayname', 'Monday'),
     ('width', 640),
     ('year', 0)]

    >>> info.set_source({'File':'/some/new/image.png', 'Pil': image2,
    ...     'Pexif':image2}, locals={'index':3, 'folderindex':5})
    >>> sorted(info.dump().items())
    [('####', '"%04d"%'), ('folderindex', 5), ('index', 3)]
    >>> info['##folderindex']
    '05'
    >>> info['root']
    u'/some'
    >>> d = info.dump('mode','height')
    >>> pprint.pprint(sorted(d.items()))
    [('##', '"%02d"%'),
     ('####', '"%04d"%'),
     ('Pil_Size', (1, 2)),
     ('folderindex', 5),
     ('foldername', u'new'),
     ('height', 2),
     ('index', 3),
     ('mode', 'L'),
     ('root', u'/some'),
     ('width', 1)]
    >>> type(d) == DumpInfo
    True
    """

    def __init__(self, locals=None, globals=GLOBALS):
        """Create an InfoEval instance. The locals get dumped, but the
        globals not with the :method:`dump` method.

        :param globals: globals that have not to be extracted (e.g. index)
        :type globals: dict
        :param locals: locals that have not to be extracted (e.g. index)
        :type locals: dict
        """
        if globals:
            self.globals = globals
        else:
            self.globals = {}
        if locals:
            self.locals = locals
        else:
            self.locals = {}
        self.list = []

    def __getitem__(self, expr):
        #return expr
        return eval(RE_PY_VAR.sub(self._regex_py_var_replace, expr),
            self.globals, self.locals)

    def _regex_py_var_replace(self, match):
        var = match.group('value')
        if var is None:
            group = 'format'
            var = match.group('format')
        else:
            group = 'value'
        #look in the info list
        for info in self.list:
            if info.provides(var):
                all = match.group(0)
                if group == 'format':
                    return info[var]
                s0 = match.start(0)
                start, end = match.span('value')
                return all[:start-s0]+repr(info[var])+all[end-s0:]
        return var

    def provides(self, var):
        for info in self.list:
            if info.provides(var):
                return True
        return False

    def append_info(self, info):
        self.list.append(info)

    def types(self):
        return [info.type for info in self.list]

    def clear(self):
        self.list = []

    def dump(self, *vars):
        """This doesn't dump globals.
        """
        if vars:
            #load vars
            for info in self.list:
                info.enable_cache()
            for var in vars:
                self[var]
            for info in self.list:
                info.disable_cache()
        d = self.locals.copy()
        for info in self.list:
            d.update(info.dict)
        return DumpInfo(d)

    def set_source(self, d, locals=None):
        """Set source of an info from the collection. Raises an
        ``UnknownTypeError`` in case an unknown type is given.

        :param d: dictionary {type: source} or type
        :type d: dict/str
        :param locals: locals that have not to be extracted (e.g. index)
        :type locals: dict
        """
        if not(locals is None):
            self.locals = locals
        for info in self.list:
            if info.type in d:
                info.set_source(d[info.type])

    def extract_all(self):
        """"""
        for info in self.list:
            info.extract_all()


class DumpInfo(dict):
    """Dictionary like object which tracks changes.

    >>> d = DumpInfo({'hello': 'world'})
    >>> d['foo'] = 'bar'
    >>> d.changed
    ['foo']
    """

    def __init__(self, d=None):
        """:param d: initial dictionary or None
        :type d: dict/None"""
        if d:
            self.update(d)
        self.changed = []

    def __setitem__(self, var, value):
        """Sets the item of the dictionary and add var to the ``changed``
        list

        :param var: variable
        :type var: string
        :param value: any kind of value
        """
        super(DumpInfo, self).__setitem__(var, value)
        self.changed.append(var)
