# Copyright (C) 2007-2008 www.stani.be
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/

# Follows PEP8

import os
import re
import pyexiv2

BROKEN = 'Exif[.]Canon'
RE_BROKEN = re.compile(BROKEN)
ISSUES = 'Saving metadata to %s caused following issues:\n'
FAILED = '''
Failed to save metadata to %s:\npyexiv2: %s
Trying again by ignoring tags with following pattern:\n%s\n'''

#info taken from http://dev.exiv2.org/wiki/exiv2/Supported_image_formats
READ_EXIF = ['JPEG', 'EXV', 'CR2', 'CRW', 'MRW', 'TIFF', 'DNG', 'NEF', 'PEF',
    'ARW', 'SR2', 'ORF', 'RW2', 'RAF', 'PSD', 'JP2']

WRITE_EXIF = ['JPEG', 'EXV', 'CRW', 'TIFF', 'DNG', 'NEF', 'PEF',
    'PSD', 'JP2']

READ_IPTC = ['JPEG', 'EXV', 'CR2', 'MRW', 'TIFF', 'DNG', 'NEF', 'PEF',
    'ARW', 'SR2', 'ORF', 'RW2', 'RAF', 'PSD', 'JP2']

WRITE_IPTC = ['JPEG', 'EXV', 'CR2', 'MRW', 'TIFF', 'DNG', 'NEF', 'PEF',
    'ARW', 'SR2', 'ORF', 'RW2', 'RAF', 'PSD', 'JP2']

READ_COMMENT = ['JPEG', 'EXV', 'CRW']

WRITE_COMMENT = ['JPEG', 'EXV', 'CRW']


def is_readable(format):
    return not format or format in READ_EXIF+READ_IPTC+READ_COMMENT


def is_writable(format):
    return not format or format in WRITE_EXIF+WRITE_IPTC+WRITE_COMMENT


def copy_metadata(source_filename, target, source_format=None,
        target_format=None, reset_orientation=None, software=None,
        update=None, thumb_data=None, source_pyexiv2_image=None):
    #if there is nothing to read or write, return immediately
    if not (is_readable(source_format) and is_writable(target_format)):
        return ''

    #read source
    source_pyexiv2_image = pyexiv2.Image(source_filename)
    source_pyexiv2_image.readMetadata()
    #correct tags
    if software:
        source_pyexiv2_image['Exif.Image.Software'] = software
    if not(reset_orientation is None):
        source_pyexiv2_image['Exif.Image.Orientation'] = reset_orientation
    for tag, value in update.items():
        source_pyexiv2_image[tag] = value
    if not source_pyexiv2_image:
        return ''

    #make two attempts to copy metadata:
    #1. normal
    #2. exclude tags which (might) break exiv2 (eg Canon tuples)

    #verify if there are tags which might break exiv2
    broken_tags = [None]
    for tag in list(source_pyexiv2_image.exifKeys())+\
        list(source_pyexiv2_image.iptcKeys()):
        if RE_BROKEN.match(tag):
            broken_tags.append(RE_BROKEN)
            break

    #copy the tags
    log = ''
    for broken_tag in broken_tags:
        #attempt to copy metadata
        try:
            warnings = _copy_metadata(source_pyexiv2_image, target,
                source_format, target_format, broken_tag, thumb_data)
            copied = True
        except Exception, message:
            copied = False

        #if metadata copied succesfully, check for warnings
        if copied:
            if warnings:
                log += ISSUES%target+warnings
            break
        log = FAILED%(target, message, BROKEN)
    return log


def _copy_metadata(source_pyexiv2_image, target, source_format=None,
        target_format=None, broken_tag=None, thumb_data=None):
    #read target
    target = pyexiv2.Image(target)
    target.readMetadata()
    warnings = []
    written = False
    #copy exif metadata
    if (not source_format or source_format in READ_EXIF) and \
        (not target_format or target_format in WRITE_EXIF):
        for tag in source_pyexiv2_image.exifKeys():
            if not(broken_tag and broken_tag.match(tag)):
                try:
                    #the following is more or less the same as
                    #target[tag] = source_pyexiv2_image[tag]
                    #but prevents conversions
                    target._Image__setExifTag(tag,
                        source_pyexiv2_image._Image__getExifTag(tag)[1])
                    written = True
                except Exception, message:
                    message = '%s: %s'%(tag, message)
                    warnings.append(message)
    #copy iptc metadata
    if (not source_format or source_format in READ_IPTC) and \
        (not target_format or target_format in WRITE_IPTC):
        for tag in source_pyexiv2_image.iptcKeys():
            try:
                target[tag] = source_pyexiv2_image[tag]
                written = True
            except Exception, message:
                message = '%s: %s'%(tag, message)
                warnings.append(message)
    #copy comment
    if (not source_format or source_format in READ_COMMENT) and \
        (not target_format or target_format in WRITE_COMMENT):
        try:
            target.setComment(source_pyexiv2_image.getComment())
            written = True
        except Exception, message:
            warnings.append(message)
    #thumbnail
    if not(thumb_data is None):
        try:
            target.setThumbnailData(thumb_data)
        except:
            pass
    #save metadata (this might rise an exception)
    if written:
        target.writeMetadata()
    return '\n'.join(warnings)


def flush(path, metadata, format=None):
    #nothing to do
    if not metadata:
        return ''
    #better try to figure out format by extension if it unknown
    if not format:
        ext = os.path.splitext(path)[-1]
        if ext:
            format = ext[1:].upper()
            if format in ['JPG', 'JPE']:
                format = 'JPEG'
            elif format == 'TIF':
                format = 'TIFF'
    if (format and not(format in WRITE_EXIF+WRITE_IPTC)):
        raise Exception('pyexiv2: writing metadata to %s is not supported'\
            %format)
    image = pyexiv2.Image(path)
    image.readMetadata()
    warnings = []
    written = False
    for tag, value in metadata.items():
        if not format or \
            (tag.startswith('Exif.') and format in WRITE_EXIF) or\
            (tag.startswith('Iptc.') and format in WRITE_IPTC):
            try:
                image[tag] = value
                written = True
            except Exception, message:
                message = '%s: %s'%(tag, message)
                warnings.append(message)
    #save metadata (this might rise an exception)
    if written:
        image.writeMetadata()
    return '\n'.join(warnings)


def example():
    import Image
    IMAGE = 'IMGA3230.JPG'
    IMAGE_PIL = 'IMGA3230_PIL.JPG'
    im = Image.open(IMAGE)
    im.save(IMAGE_PIL)
    copy_metadata(IMAGE, IMAGE_PIL)


if __name__ == '__main__':
    test()
