# Copyright (C) 2007-2008 www.stani.be
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/

# Follows PEP8

import hashlib
import os
import re
from cStringIO import StringIO
from itertools import cycle
from urllib import urlopen

import Image
import ImageDraw
import ImageEnhance

from desktop import USER_THUMBNAILS_NORMAL_FOLDER
from lib import system

CHECKBOARD = {}
THUMB_SIZE = (64, 64)
COLOR_MAP = [255] * 128 + [0] * 128


def open_image(x):
    if system.is_www_file(x):
        try:
            return WWW_CACHE[x]
        except KeyError:
            f = urlopen(x)
            im = WWW_CACHE[x] = Image.open(StringIO(f.read()))
            return im
    return Image.open(x)


class Counter:
    bytes = 0

    def write(self, data):
        self.bytes += len(data)


def get_size(im, format, **options):
    try:
        out = Counter()
        im.save(out, format, **options)
        return out.bytes
    except AttributeError:
        # fall back on full in-memory compression
        out = StringIO()
        im.save(out, format, **options)
        return len(out.getvalue())


def get_quality(im, size, format, down=0, up=100, delta=1000, options=None):
    if options is None:
        options = {}
    q = options['quality'] = (down+up)/2
    if q==down or q==up:
        return max(q, 1)
    s = get_size(im, format, **options)
    if abs(s-size)<delta:
        return q
    elif s > size:
        return get_quality(im, size, format, down, up=q, options=options)
    else:
        return get_quality(im, size, format, down=q, up=up, options=options)


def fill_background_color(image, color):
    """Fills given image with background color"""
    if image.mode == 'LA':
        image = image.convert('RGBA')
    elif image.mode != 'RGBA' and\
        not (image.mode == 'P' and 'transparency' in image.info):
        return image
    if len(color) == 4 and color[-1]!=255:
        mode = 'RGBA'
    else:
        mode = 'RGB'
    back = Image.new(mode, image.size, color)
    if (image.mode == 'P' and mode == 'RGBA'):
        image = image.convert('RGBA')
    if image.mode in ['RGBA', 'LA']:
        back.paste(image, image)
    elif image.mode == 'P':
        palette = image.getpalette()
        index = image.info['transparency']
        palette[index * 3: index * 3 + 3] = color[:3]
        image.putpalette(palette)
        del image.info['transparency']
        back = image
    else:
        back.paste(image)
    return back


def generate_layer(image, mark, method,
                   horizontal_offset, vertical_offset,
                   horizontal_justification, vertical_justification,
                   orientation, opacity):
    image = image.convert('RGBA')

    mark = open_image(mark)
    opacity /= 100.0
    mark = reduce_opacity(mark, opacity)

    layer = Image.new('RGBA', image.size, (0, 0, 0, 0))

    if method == 'Tile':
        for y in range(0, image.size[1], mark.size[1]):
            for x in range(0, image.size[0], mark.size[0]):
                layer.paste(mark, (x, y))
    elif method == 'Scale':
        # scale, but preserve the aspect ratio
        ratio = min(float(image.size[0]) / mark.size[0],
            float(image.size[1]) / mark.size[1])
        w = int(mark.size[0] * ratio)
        h = int(mark.size[1] * ratio)
        mark = mark.resize((w, h))
        layer.paste(mark, ((image.size[0] - w) / 2,
            (image.size[1] - h) / 2))
    elif method == 'By Offset':
        location = calculate_location(
            horizontal_offset, vertical_offset,
            horizontal_justification, vertical_justification,
            image.size, mark.size)

        if orientation:
            orientation_value = getattr(Image, orientation)
            mark = mark.transpose(orientation_value)

        layer.paste(mark, location)

    return layer


def reduce_opacity(im, opacity):
    """Returns an image with reduced opacity if opacity is
    within [0, 1]."""
    if opacity < 0 or opacity > 1:
        return im
    alpha = get_alpha(im)
    alpha = ImageEnhance.Brightness(alpha).enhance(opacity)
    put_alpha(im, alpha)
    return im


def calculate_location(horizontal_offset, vertical_offset,
        horizontal_justification, vertical_justification,
        canvas_size, image_size):
    canvas_width, canvas_height = canvas_size
    image_width, image_height = image_size

    # check offsets
    if horizontal_offset < 0:
        horizontal_offset += canvas_width
    if vertical_offset < 0:
        vertical_offset += canvas_height

    # check justifications
    if horizontal_justification == 'Left':
        horizontal_delta = 0
    elif horizontal_justification == 'Middle':
        horizontal_delta = -image_width / 2
    elif horizontal_justification == 'Right':
        horizontal_delta = -image_width

    if vertical_justification == 'Top':
        vertical_delta = 0
    elif vertical_justification == 'Middle':
        vertical_delta = -image_height / 2
    elif vertical_justification == 'Bottom':
        vertical_delta = -image_height

    return horizontal_offset + horizontal_delta, \
        vertical_offset + vertical_delta


def checkboard(size, delta=8, fg=(128, 128, 128), bg=(204, 204, 204)):
    "Draw an n x n checkboard"
    if not (size in CHECKBOARD):
        dim = max(size)
        n = int(dim/delta)+1 #FIXME: now acts like square->nx, ny

        def sq_start(i):
            "Return the x/y start coord of the square at column/row i."
            return i * delta

        def square(i, j):
            "Return the square corners"
            return map(sq_start, [i, j, i + 1, j + 1])

        image = Image.new("RGB", size, bg)
        draw_square = ImageDraw.Draw(image).rectangle
        squares = (square(i, j)
           for i_start, j in zip(cycle((0, 1)), range(n))
           for i in range(i_start, n, 2))
        for sq in squares:
            draw_square(sq, fill=fg)
        CHECKBOARD[size] = image
    return CHECKBOARD[size].copy()


def thumb_from_file(filename, thumb_size=THUMB_SIZE,
        filter=Image.ANTIALIAS, write_cache=True, image=None):
    # can we use the freedesktop cache in ~/.thumbnails/normal?
    cache = USER_THUMBNAILS_NORMAL_FOLDER and \
        (filename.startswith('/') or filename.startswith('file:///'))
    # is the thumbnail small enough for the cache?
    use_cache = thumb_size[0]<129 and thumb_size[1]<129 and cache
    write_cache = use_cache and write_cache
    # retrieve the filename of the cached thumbnail if desirable
    if use_cache:
        uri = filename
        if not filename.startswith('file:///'):
            uri = 'file://'+uri
        digest = hashlib.md5(uri).hexdigest()
        t = os.path.join(USER_THUMBNAILS_NORMAL_FOLDER, digest+'.png')
        if os.path.isfile(t):
            image = Image.open(t)
        elif write_cache:
            if image is None:
                image = Image.open(filename)
            else:
                image = image.copy()
            if not image.mode in ('RGB', 'RGBA'):
                image = image.convert('RGB')
            image.thumbnail((128, 128), Image.ANTIALIAS)
            image.save(t)
    else:
        image = Image.open(filename)
    return thumb(image, thumb_size, filter)


def thumb(image, thumb_size=THUMB_SIZE, filter=Image.ANTIALIAS):
    image = image.copy()
    image.thumbnail(thumb_size, filter)
    if image.mode == 'RGBA':
        image_bg = checkboard(image.size)
        image_bg.paste(image, (0, 0), image)
        return image_bg
    else:
        return image

####################################
####    PIL helper functions    ####
####################################

def has_alpha(image):
    """Checks if the image has an alpha band.
    i.e. the image mode is either RGBA or LA.
    The transparency in the P mode doesn't count as an alpha band

    :param image: the image to check
    :type image: PIL image object
    :returns: True or False
    :rtype: boolean
    """
    return image.mode in ['RGBA', 'LA']


def has_transparency(image):
    """Checks if the image has transparency.
    The image has an alpha band or a P mode with transparency.

    :param image: the image to check
    :type image: PIL image object
    :returns: True or False
    :rtype: boolean
    """
    return (image.mode == 'P' and 'transparency' in image.info) or\
            has_alpha(image)


def get_alpha(image):
    """Gets the image alpha band. Can handles P mode images with transpareny.
    Returns a band with all values set to 255 if no alpha band exists.

    :param image: input image
    :type image: PIL image object
    :returns: alpha as a band
    :rtype: single band image object
    """
    if has_alpha(image):
        return image.split()[-1]
    if image.mode == 'P' and 'transparency' in image.info:
        return image.convert('LA').split()[-1]
    # No alpha layer, create one.
    return Image.new('L', image.size, 255)


def get_format_data(image, format):
    f = StringIO()
    image.save(f, format)
    return f.getvalue()


def get_thumb_data(image, format, size=(160, 160)):
    return get_format_data(thumb(image, size), format)


def get_reverse_transposition(transposition):
    if transposition == Image.ROTATE_90:
        return Image.ROTATE_270
    elif transposition == Image.ROTATE_270:
        return Image.ROTATE_90
    return transposition


def put_alpha(image, alpha):
    """Copies the given band to the alpha layer of the given image.

    :param image: input image
    :type image: PIL image object
    :param alpha: the alpha band to copy
    :type alpha: single band image object
    """
    if image.mode in ['CMYK', 'YCbCr', 'P']:
        image = image.convert('RGBA')
    elif image.mode in ['1', 'F']:
        image = image.convert('LA')
    image.putalpha(alpha)


def remove_alpha(image):
    """Returns a copy of the image after removing the alpha band or
    transparency

    :param image: input image
    :type image: PIL image object
    :returns: the input image after removing the alpha band or transparency
    :rtype: PIL image object
    """
    if image.mode == 'RGBA':
        return image.convert('RGB')
    if image.mode == 'LA':
        return image.convert('L')
    if image.mode == 'P' and 'transparency' in image.info:
        return image.convert('RGB')
    return image


def convert_safe_mode(image):
    """Converts image into a processing-safe mode.

    :param image: input image
    :type image: PIL image object
    :returns: the converted image
    :rtype: PIL image object
    """
    if image.mode in ['1', 'F']:
        return image.convert('L')
    if image.mode == 'P' and 'transparency' in image.info:
        return image.convert('RGBA')
    if image.mode in ['P', 'YCbCr', 'CMYK', 'RGBX']:
        return image.convert('RGB')
    return image


def auto_crop(image):
    """Crops all transparent or black background from the image
    :param image: input image
    :type image: PIL image object
    :returns: the cropped image
    :rtype: PIL image object
    """

    alpha = get_alpha(image)
    box = alpha.getbbox()
    return convert_safe_mode(image).crop(box)


def convert(image, mode, *args, **keyw):
    """Returns a converted copy of an image

    :param image: input image
    :type image: PIL image object
    :param mode: the new mode
    :type mode: string
    :param args: extra options
    :type args: tuple of values
    :param keyw: extra keyword options
    :type keyw: dictionary of options
    :returns: the converted image
    :rtype: PIL image object
    """
    if mode == image.mode:
        return image
    if mode == 'P':
        if image.mode in ['1', 'F']:
            return image.convert('L').convert(mode, *args, **keyw)
        if image.mode in ['RGBA', 'LA']:
            alpha = image.split()[-1]
            output = image.convert('RGB').convert(
                mode, colors=255, *args, **keyw)
            output.paste(
                255, alpha.point(COLOR_MAP))
            return output
        return image.convert('RGB').convert(mode, *args, **keyw)
    if has_transparency(image) and (not mode in ['RGBA', 'LA']):
        if image.mode == 'P':
            image = image.convert('RGBA')
        image = fill_background_color(image, (255, 255, 255, 255))
        return image.convert(mode, *args, **keyw)
    return image.convert(mode, *args, **keyw)


def example():
    filename = '/home/stani/sync/Desktop/IMGA3345.JPG'
    im = Image.open(filename)
    q = get_quality(im, 300000, "JPEG")
    im.save(filename.replace('.jpg', '_sized.jpg'))


if __name__ == '__main__':
    test()
