# Copyright (C) 2009 www.stani.be
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/

# Follows PEP8

import os
import re
import sys

import Image

import system


def open_image(filename, command, app, extension='png', temp_ext=None):
    """Open with an external command (such as Inkscape, dcraw, imagemagick).

    :param filename: filename
    :type filename: string
    :param command: conversion command with optional temp file interpolation
    :type command: string
    :param extension: file type
    :type extension: string
    :param temp_ext:

        if a temp file can not be specified to the command (eg dcraw),
        give the file extension of the command output

    :type temp_ext: string
    """
    if temp_ext:
        # eg dcraw
        temp = os.path.splitext(filename)[0]+'.'+temp_ext
    else:
        # imagemagick, ...
        temp = system.tempfile_name()+'.'+extension
        command = command%temp
    output, error = system.shell(command)
    if os.path.exists(temp):
        image = Image.open(temp)
        image.load() # otherwise temp file can't be deleted
        os.remove(temp)
        image.info['open'] = app
        return image
    message = _('Could not open image with %s.')%app + \
        '\n\n%s: %s\n\n%s: %s\n\n%s: %s'% \
        (_('Command'), command, _('Output'), output, _('Error'), error)
    raise IOError(message)


#libtiff

TIFFINFO = system.find_exe("tiffinfo")
TIFFCP = system.find_exe("tiffcp")

if TIFFINFO and TIFFCP:

    RE_TIFF_FIELD = re.compile('\s+(.*?):\s+(.*?)\n')
    RE_TIFF_FIELD_IMAGE = re.compile(' Image (.*?):\s+(.*?)$')

    TIFF_COMPRESSION = {
        'CCITT Group 3': 'g3',
        'CCITT Group 4': 'g4',
        'Deflate': 'zip',
        'JPEG': 'jpeg',
        'LZW': 'lzw',
        'PackBits': 'packbits',
        'None': 'none',
    }

    TIFF_COMPRESSION_TYPES = TIFF_COMPRESSION.values()
    TIFF_COMPRESSION_TYPES.sort()

    def get_info_libtiff(filename):
        """Get tiff info of a file with ``tiffinfo``, which needs to be
        installed on your system.

        :param filename: name of tiff image file
        :type filename: string
        :returns: info about the file
        :rtype: dict
        """
        result = {}

        def set(key, value):
            key = 'libtiff.'+key.lower().replace(' ', '.')
            if not (key in result):
                result[key] = value

        source, err = system.shell('%s %s'%(TIFFINFO, filename))
        for match in RE_TIFF_FIELD.finditer(source):
            value = match.group(2)
            again = RE_TIFF_FIELD_IMAGE.search(value)
            if again:
                set('Image %s'%again.group(1), again.group(2))
                set(match.group(1), value[:again.start()])
            else:
                set(match.group(1), value)
        if not result:
            raise IOError('Not a TIFF or MDI file, bad magic number.')
        result['compression'] = \
            TIFF_COMPRESSION[result['libtiff.compression.scheme']]
        return result

    def open_libtiff(filename):
        """Opens a tiff file with ``tiffcp``, which needs to be installed
        on your system.

        :param filename: name of tiff image file
        :type filename: string
        :returns: PIL image
        :rtype: Image.Image
        """
        # get info
        info = get_info_libtiff(filename)
        # extract
        temp = system.tempfile_name()
        system.shell('%s -c none -r -1 %s %s'%(TIFFCP, filename, temp))
        if os.path.exists(temp):
            image = Image.open(temp)
            image.load() # otherwise temp file can't be deleted
            os.remove(temp)
            image.info.update(info)
            image.info['open'] = 'libtiff'
        else:
            raise IOError('Could not extract tiff image with tiffcp.')
        return image

    def save_libtiff(image, filename, compression=None, **options):
        """Saves a tiff compressed file with tiffcp.

        :param image: PIL image
        :type image: Image.Image
        :param filename: name of tiff image file
        :type filename: string
        :param compression: g3, g4, jpeg, lzw, tiff_lzw
        :type compression: string
        :returns: log message
        :rtype: string
        """
        if compression in ['raw', 'none']:
            image.save(filename, 'tiff', **options)
            return ''
        if compression is None:
            compression = image.info['compression']
        elif compression in ['g3', 'g4'] and image.mode != '1':
            image = image.convert('1')
        if compression == 'jpeg':
            if image.mode == 'RGBA':
                image = image.convert('RGB')
            option = ' -r 16'
        elif compression == 'tiff_lzw':
            compression = 'lzw'
            option = ''
        else:
            option = ''
        temp = system.tempfile_name()
        image.save(temp, 'tiff', **options)
        input = '%s %s -c %s%s %s' \
            %(TIFFCP, temp, compression, option, filename)
        out, err = system.shell(input)
        os.remove(temp)
        if out or err:
            return 'Subprocess "tiffcp"\ninput:\n%s\noutput:\n%s%s\n'\
                %(input, out, err)
        return ''

else:
    open_libtiff = save_libtiff = get_info_libtiff = None


# inkscape

INKSCAPE = system.find_exe('inkscape')

if INKSCAPE:

    def open_inkscape(filename):
        """Open an Inkscape file."""
        command = '%s "%s" -e "%%s"'%(INKSCAPE, filename)
        return open_image(filename, command, 'inkscape')

else:

    open_inkscape = None

# imagemagick

IMAGEMAGICK_CONVERT = system.find_exe('convert')

if IMAGEMAGICK_CONVERT:

    def open_imagemagick(filename):
        """Open an image with Imagemagick."""
        command = '%s "%s" -flatten "%%s"'%(IMAGEMAGICK_CONVERT, filename)
        return open_image(filename, command, 'imagemagick')

else:

    open_imagemagick = None


# xcf tools (gimp)

XCF2PNG = system.find_exe('xcf2png')

if XCF2PNG:

    def open_xcf(filename):
        """Open a gimp file."""
        command = '%s "%s" -o "%%s"'%(XCF2PNG, filename)
        return open_image(filename, command, 'xcf2png')

else:

    open_xcf = None


# dcraw

DCRAW = system.find_exe('dcraw')

if DCRAW:

    def open_dcraw(filename):
        """Open a camera raw image file."""
        command = '%s -w "%s"'%(DCRAW, filename)
        return open_image(filename, command, 'xcf2png',
            temp_ext='ppm')

else:

    open_dcraw = None

# register methods

# IMPORTANT: the order of registering is important, the method
# which is first registered gets more priority

RAW_EXTENSIONS = ['arw', 'cr2', 'crw', 'dcr', 'nef', 'orf', 'x3f']

WITHOUT_PIL = system.MethodRegister()
WITHOUT_PIL.register(['xcf'], open_xcf)
WITHOUT_PIL.register(RAW_EXTENSIONS, open_dcraw)
WITHOUT_PIL.register(['svg', 'svgz'], open_inkscape)
WITHOUT_PIL.register(['eps', 'psd', 'pdf', 'svg', 'svgz', 'wmf', 'xcf'],
    open_imagemagick)

# This is for file formats which PIL can read, but not all subformats
# For example: compressed tiff files

ENHANCE_PIL = system.MethodRegister()
ENHANCE_PIL.register(['tiff'], open_libtiff)

if __name__ == '__main__':
    print('Libtiff-tools installed: %s'%system.init())
    print TIFF_COMPRESSION_TYPES
    filename = '/home/stani/Downloads/0009.tif'
    image = open(filename)
    print(image.info)
    save(image.rotate(10), filename+'.tif', compression='g4')
