#!/usr/bin/python

import glob
import os
import sys
import time
import logging
import optparse
from subprocess import Popen, PIPE
import pprint

DISABLE_ACTIONS = ['save', 'save tags', 'rename', 'blender']

#import gettext
#gettext.install('test')

def get_actions(path):
    """Get a dictionary that maps action names to action classes"""
    action_names = [
        os.path.basename(action_file).partition('.')[0]
        for action_file in glob.glob(os.path.join(path, '*.py'))
    ]
    return dict(
        (name, __import__('phatch.actions.%s' % name, name, fromlist=['phatch.actions']).Action())
        for name in action_names
        if name != '__init__'
    )


def get_actionlist(actions):
    """Get actionlist structure from a list of actions"""
    data = [action.dump() for action in actions]
    return {'description': '', 'actions': data}


def build_library_actionlists(library_path, in_path, out_path):
    """Build library actionlists"""
    if not os.path.exists(in_path):
        os.makedirs(in_path)
    result = []
    actionlists = [
        (os.path.basename(actionlist_file).partition('.')[0], eval(open(actionlist_file).read()))
        for actionlist_file in glob.glob(os.path.join(library_path, '*.phatch'))
    ]
    for name, actionlist in actionlists:
        save_actions = [
            action
            for action in actionlist['actions']
            if action['label'] == 'Save'
        ]
        for save_action in save_actions:
            save_action['fields']['In'] = '%s/<subfolder>' % out_path
            save_action['fields']['File Name'] = '<filename>_%s' % name
        result.append(write_file(os.path.join(
                in_path,
                '%s.phatch' % name
            ),
            actionlist
        ))
    return result


def build_actionlists(actions, in_path, out_path):
    """Build actionlists"""
    actionlists = []
    if not os.path.exists(in_path):
        os.makedirs(in_path)
    save_action = actions['save']
    save_action.set_field(
        'In',
        '%s/<subfolder>' % out_path
    )
    for name, action in actions.iteritems():
        if name in DISABLE_ACTIONS:
            continue
        choices = []
        possible_choices(action, choices)
        for choice in choices:
            filename = [name]
            for fname, fvalue in choice.iteritems():
                action.set_field(fname, fvalue)
                filename.append('%s=%s' % (fname.lower(), str(fvalue).lower()))
            filename = '_'.join(filename)
            save_action.set_field(
                'File Name',
                '<filename>_%s' % filename
            )
            actionlists.append(write_file(
                os.path.join(
                    in_path,
                    '%s.phatch' % filename
                ),
                get_actionlist([action, save_action])
            ))
    return actionlists


def possible_choices(action, choices):
    """Generate all possible action choices"""
    if hasattr(action, 'get_relevant_field_labels'):
        relevant = action.get_relevant_field_labels()
    else:
        relevant = action._fields.keys()
    choice_fields = dict(
        (name, field) for name, field in action._fields.iteritems()
        if isinstance(field, (action.BooleanField, action.ChoiceField))
        and not isinstance(field, (
            action.ImageFilterField,
            action.ImageResampleField,
            action.ImageResampleAutoField)
        )
        and name != '__enabled__'
        and name in relevant
    )
    choice = dict(
        (fname, field.get())
        for fname, field in choice_fields.iteritems()
    )
    if choice in choices:
        return
    choices.append(choice)
    for fname, field in choice_fields.iteritems():
        default = field.get()
        if isinstance(field, action.BooleanField):
            options = [True, False]
        else:
            options = field.choices
        for option in options:
            if option == default:
                continue
            action.set_field(fname, option)
            possible_choices(action, choices)


def write_file(path, data):
    """Write a file to disk"""
    f = open(
        path,
        'w'
    )
    f.write(pprint.pformat(data))
    f.close()
    return path


def execute(input, actionlist):
    """Execute the actionlist in phatch"""
    output, error = Popen(
        ['python', '../phatch/phatch.py', '-c', actionlist, input],
        stdout=open('logs', 'a'), stderr=open('logs', 'a')
    ).communicate()
    if output:
        logging.info(
            'Executing actionlist: %s generated the following output:\n%s'
            % (actionlist, output)
        )
    if error:
        logging.error(
            'Executing actionlist: %s generated the following error:\n%s'
            % (actionlist, error)
        )


if __name__ == '__main__':
    # Hack to be able to import phatch
    sys.path.append('..')

    # Logging
    logging.basicConfig(
        filename='logs',
        level=logging.DEBUG,
        format='%(levelname)s: %(message)s'
    )

    # Option parser
    parser = optparse.OptionParser()
    parser.add_option(
        '-i', '--input',
        default='images/source',
        help='Image input folder [default: %default]'
    )
    parser.add_option(
        '-o', '--output',
        default='images/output',
        help='Image output folder [default: %default]'
    )
    parser.add_option(
        '-a', '--actionlists',
        action='store_true',
        default=False,
        help='Generate actionlists only'
    )
    parser.add_option(
        '-l', '--library',
        action='store_true',
        default=False,
        help='Run library actionlists tests only'
    )
    options, args = parser.parse_args()
    options.output = os.path.abspath(options.output)
    start = time.time()

    # Getting actions
    actions = get_actions('../phatch/actions')

    # Generating actionlists
    try:
        print 'Generating library actionlists...',
        actionlists = build_library_actionlists(
            '../data/actionlists',
            'test_actionlists',
            options.output
        )
        print 'Done'

        if not options.library:
            sys.stdout.write('Generating all possible actionlists...')
            sys.stdout.flush()
            actionlists.extend(
                build_actionlists(actions, 'test_actionlists', '<root>/output')
            )
            print 'Done'
        if not options.actionlists:
            total = len(actionlists)
            for i, name in enumerate(actionlists):
                sys.stdout.write(' Running %s/%s\r' % (i, total))
                sys.stdout.flush()
                execute(options.input, name)
    except KeyboardInterrupt:
        print '\nStopped'
    print 'Execution took %.2f seconds' % (time.time() - start)


